#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test suite for the response handling.

:copyright:
    Lion Krischer (krischer@geophysik.uni-muenchen.de), 2013
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import warnings
from copy import deepcopy
from math import pi

import numpy as np
import pytest
import scipy.interpolate

from obspy import UTCDateTime, read_inventory
from obspy.core.inventory.response import (
    _pitick2latex, PolesZerosResponseStage, PolynomialResponseStage, Response,
    ResponseListResponseStage, ResponseListElement, InstrumentSensitivity)
from obspy.core.util.base import CatchAndAssertWarnings
from obspy.core.util.misc import CatchOutput
from obspy.core.util.obspy_types import ComplexWithUncertainties
from obspy.signal.invsim import evalresp
from obspy.io.xseed import Parser


@pytest.mark.usefixtures('ignore_numpy_errors')
class TestResponse:
    """
    Tests the for :class:`~obspy.core.inventory.response.Response` class.
    """
    def test_evalresp_with_output_from_seed(self, datapath):
        """
        The StationXML file has been converted to SEED with the help of a tool
        provided by IRIS:

        https://seiscode.iris.washington.edu/projects/stationxml-converter
        """
        t_samp = 0.05
        nfft = 16384

        # Test for different output units.
        units = ["DISP", "VEL", "ACC"]
        filenames = ["IRIS_single_channel_with_response", "XM.05", "AU.MEEK"]

        for filename in filenames:
            path = datapath / filename
            xml_filename = str(path) + ".xml"
            seed_filename = str(path) + ".seed"

            p = Parser(seed_filename)

            # older systems don't like an end date in the year 2599
            t_ = UTCDateTime(2030, 1, 1)
            if p.blockettes[50][0].end_effective_date > t_:
                p.blockettes[50][0].end_effective_date = None
            if p.blockettes[52][0].end_date > t_:
                p.blockettes[52][0].end_date = None

            resp_filename = p.get_resp()[0][-1]

            inv = read_inventory(xml_filename)

            network = inv[0].code
            station = inv[0][0].code
            location = inv[0][0][0].location_code
            channel = inv[0][0][0].code
            date = inv[0][0][0].start_date

            for unit in units:
                resp_filename.seek(0, 0)

                seed_response, seed_freq = evalresp(
                    t_samp, nfft, resp_filename, date=date, station=station,
                    channel=channel, network=network, locid=location,
                    units=unit, freq=True)

                xml_response, xml_freq = \
                    inv[0][0][0].response.get_evalresp_response(t_samp, nfft,
                                                                output=unit)

                assert np.allclose(seed_freq, xml_freq, rtol=1E-5)
                assert np.allclose(seed_response, xml_response, rtol=1E-5)

                # also test getting response for a set of discrete frequencies
                indices = (-2, 0, -1, 1, 2, 20, -30, -100)
                freqs = [seed_freq[i_] for i_ in indices]
                response = inv[0][0][0].response
                got = response.get_evalresp_response_for_frequencies(
                    freqs, output=unit)
                expected = [seed_response[i_] for i_ in indices]
                np.testing.assert_allclose(got, expected, rtol=1E-5)

    def test_calculate_normalization_factor(self):
        """
        Test calculation of the normalization factor for pole–zero stages given
        in both radians/second and hertz. Streckeisen STS-1 poles and zeros and
        baseline values ("A0 normalization") are taken from:
        https://docs.fdsn.org/projects/stationxml/en/latest/response.html#stage-1-the-analog-sensor
        """
        _paz_sts1_kwargs = dict(
            stage_sequence_number=1,
            stage_gain=2400,
            stage_gain_frequency=0.02,
            input_units='M/S',
            output_units='V',
            normalization_frequency=0.02,
        )
        # (1) Test for radians/second — use normalization_factor=1.0 (default)
        paz_sts1_rads = PolesZerosResponseStage(
            pz_transfer_function_type='LAPLACE (RADIANS/SECOND)',
            poles=[
                -0.01234 + 0.01234j,
                -0.01234 - 0.01234j,
                -39.18 + 49.12j,
                -39.18 - 49.12j,
            ],
            zeros=[0j, 0j],
            **_paz_sts1_kwargs,
        )
        np.testing.assert_allclose(  # Calculation is run below
            paz_sts1_rads.calculate_normalization_factor(), 3948.58, rtol=1e-5
        )
        # (2) Test for hertz — use normalization_factor=None to calculate it
        paz_sts1_hz = PolesZerosResponseStage(
            pz_transfer_function_type='LAPLACE (HERTZ)',
            poles=[
                -0.0019639 + 0.0019639j,
                -0.0019639 - 0.0019639j,
                -6.2357 + 7.8177j,
                -6.2357 - 7.8177j,
            ],
            zeros=[0j, 0j],
            normalization_factor=None,  # Calculated at initialization time
            **_paz_sts1_kwargs,
        )
        np.testing.assert_allclose(  # Just accessing the calculated attribute
            paz_sts1_hz.normalization_factor, 100.01869, rtol=1e-5
        )
        # (3) Test for an unsupported transfer function type (use dummy paz)
        paz_dummy = PolesZerosResponseStage(
            pz_transfer_function_type='DIGITAL (Z-TRANSFORM)',
            poles=[],
            zeros=[],
            **_paz_sts1_kwargs,
        )
        with pytest.raises(NotImplementedError):
            paz_dummy.calculate_normalization_factor()

    def test_pitick2latex(self):
        assert _pitick2latex(3 * pi / 2) == r'$\frac{3\pi}{2}$'
        assert _pitick2latex(2 * pi / 2) == r'$\pi$'
        assert _pitick2latex(1 * pi / 2) == r'$\frac{\pi}{2}$'
        assert _pitick2latex(0 * pi / 2) == r'$0$'
        assert _pitick2latex(-1 * pi / 2) == r'$-\frac{\pi}{2}$'
        assert _pitick2latex(-2 * pi / 2) == r'$-\pi$'
        assert _pitick2latex(0.5) == r'0.500'
        assert _pitick2latex(3 * pi + 0.01) == r'9.43'
        assert _pitick2latex(30 * pi + 0.01) == r'94.3'
        assert _pitick2latex(300 * pi + 0.01) == r'942.'
        assert _pitick2latex(3000 * pi + 0.01) == r'9.42e+03'

    def test_response_plot(self, image_path):
        """
        Tests the response plot.
        """
        resp = read_inventory()[0][0][0].response
        with CatchAndAssertWarnings():
            resp.plot(0.001, output="VEL", start_stage=1, end_stage=3,
                      outfile=image_path)

    def test_response_plot_degrees(self, image_path):
        """
        Tests the response plot in degrees.
        """
        resp = read_inventory()[0][0][0].response
        with CatchAndAssertWarnings():
            resp.plot(0.001, output="VEL", start_stage=1, end_stage=3,
                      plot_degrees=True, outfile=image_path)

    def test_segfault_after_error_handling(self, testdata):
        """
        Many functions in evalresp call `error_return()` which uses longjmp()
        to jump to some previously set state.

        ObsPy calls some evalresp functions directly so evalresp cannot call
        setjmp(). In that case longjmp() jumps to an undefined location, most
        likely resulting in a segfault.

        This test tests a workaround for this issue.

        As long as it does not segfault the test is doing alright.
        """
        inv = read_inventory(testdata["TM.SKLT.__.BHZ_faulty_response.xml"])

        t_samp = 0.05
        nfft = 256

        resp = inv[0][0][0].response
        with CatchOutput():
            with pytest.raises(ValueError):
                resp.get_evalresp_response(t_samp, nfft, output="DISP")

    def test_custom_types_init(self):
        """
        Test initializations that involve custom decimal types like
        `ComplexWithUncertainties`.
        """
        # initializing poles / zeros from native types should work
        poles = [1 + 1j, 1, 1j]
        zeros = [2 + 3j, 2, 3j]
        stage = PolesZerosResponseStage(
            1, 1, 1, "", "", "LAPLACE (HERTZ)", 1, zeros, poles)
        assert isinstance(stage.zeros[0], ComplexWithUncertainties)
        assert isinstance(stage.poles[0], ComplexWithUncertainties)
        assert stage.poles == poles
        assert stage.zeros == zeros

    def test_response_list_stage(self, testdata):
        """
        This is quite rare but it happens.
        """
        inv = read_inventory(testdata["IM_IL31__BHZ.xml"])

        sampling_rate = 40.0
        t_samp = 1.0 / sampling_rate
        nfft = 100

        with CatchAndAssertWarnings():
            cpx_response, freq = inv[0][0][0].response.get_evalresp_response(
                t_samp=t_samp, nfft=nfft, output="VEL", start_stage=None,
                end_stage=None)

        # Cut of the zero frequency.
        cpx_response = cpx_response[1:]

        amp = np.abs(cpx_response)
        phase = np.angle(cpx_response)
        freq = freq[1:]

        # The expected output goes from 1 to 20 Hz - its somehow really hard
        # to get evalresp to produce results for the desired frequencies so
        # I just gave up on it.
        exp_f, exp_amp, exp_ph = np.loadtxt(
            testdata["expected_response_IM_IL31__BHZ.txt"]).T
        # Interpolate.
        exp_amp = scipy.interpolate.InterpolatedUnivariateSpline(
            exp_f, exp_amp, k=3)(freq)
        exp_ph = scipy.interpolate.InterpolatedUnivariateSpline(
            exp_f, exp_ph, k=3)(freq)
        exp_ph = np.deg2rad(exp_ph)

        # The output is not exactle the same as ObsPy performs a different
        # but visually quite a bit better interpolation.
        np.testing.assert_allclose(amp, exp_amp, rtol=1E-3)
        np.testing.assert_allclose(phase, exp_ph, rtol=1E-3)

    def test_response_with_no_units_in_stage_1(self, testdata):
        """
        ObsPy has some heuristics to deal with this particular degenerate case.
        Test it here.
        """
        inv = read_inventory(testdata["stationxml_no_units_in_stage_1.xml"])
        r = inv[0][0][0].response

        # The units should already have been fixed from reading the StationXML
        # files...
        assert r.response_stages[0].input_units == "M/S"
        assert r.response_stages[0].input_units_description == \
            "Meters per second"
        assert r.response_stages[0].output_units == "V"
        assert r.response_stages[0].output_units_description == "VOLTS"

        # We have to set the units to None here as there is some other logic in
        # reading the StationXML files that sets them based on other units...
        r.response_stages[0].input_units = None
        r.response_stages[0].output_units = None

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            out = r.get_evalresp_response_for_frequencies(
                np.array([0.5, 1.0, 2.0]), output="DISP")

        assert len(w) == 2
        assert w[0].message.args[0] == \
            "Set the input units of stage 1 to the overall input units."
        assert w[1].message.args[0] == \
            "Set the output units of stage 1 to the input units of stage 2."

        # Values compared to evalresp output from RESP file - might not be
        # right but it does guarantee that ObsPy behaves like evalresp - be
        # that a good thing or a bad thing.
        np.testing.assert_allclose(
            out, [0 + 9869.2911771081963j, 0 + 19738.582354216393j,
                  0 + 39477.164708432785j])

    def test_resp_from_paz_setting_zeros_poles_and_sensitivity(self):
        poles = [1 + 2j, 1 - 2j, 2 + 1j, 2 - 1j]
        zeros = [0, 0, 5]
        sensitivity = 1201. * (2 ** 26 / 40.)
        # a0 normalization factor at 1Hz for these values
        normalization = np.prod(2 * pi * 1j - np.array(zeros))
        normalization /= np.prod(2 * pi * 1j - np.array(poles))
        normalization = np.abs(normalization)
        resp = Response.from_paz(zeros, poles, sensitivity)
        r_zeros = resp.response_stages[0].zeros
        r_poles = resp.response_stages[0].poles
        r_stage_gain = resp.response_stages[0].stage_gain
        r_sens = resp.instrument_sensitivity.value
        np.testing.assert_array_equal(r_zeros, zeros)
        np.testing.assert_array_equal(r_poles, poles)
        np.testing.assert_equal(r_stage_gain, sensitivity)
        np.testing.assert_allclose(r_sens / normalization, sensitivity, atol=0,
                                   rtol=1e-8)

    def test_resp_from_paz_loading_vs_evalresp(self, testdata):
        zeros = [0., 0.]
        poles = [-8.443 + 1.443j, -8.443 - 1.443j]
        filename = testdata['RESP.XX.NS306..SHZ.GS13.1.2180']
        resp_er = read_inventory(filename)[0][0][0].response
        loaded_resp = resp_er.get_evalresp_response(.1, 2**6, output='VEL')
        # The optional kwargs are the same as those being set in the RESP file.
        resp = Response.from_paz(zeros=zeros, poles=poles, stage_gain=22.0,
                                 stage_gain_frequency=5.0,
                                 normalization_frequency=5.0,
                                 normalization_factor=1.070401)
        paz_resp = resp.get_evalresp_response(.1, 2**6, output='VEL')
        np.testing.assert_allclose(paz_resp, loaded_resp)

    def test_str_method_of_the_polynomial_response_stage(self):
        # First with gain and gain frequency.
        assert str(PolynomialResponseStage(
            stage_sequence_number=2,
            stage_gain=12345.0,
            stage_gain_frequency=1.0,
            input_units="PA",
            input_units_description="Pascal",
            output_units="COUNTS",
            output_units_description="digital_counts",
            frequency_lower_bound=1.0,
            frequency_upper_bound=2.0,
            approximation_lower_bound=3.0,
            approximation_upper_bound=4.0,
            maximum_error=1.5,
            coefficients=[1.0, 2.0, 3.0],
            approximation_type="MACLAURIN",
            decimation_input_sample_rate=1.0,
            decimation_factor=2,
            decimation_offset=3.0,
            decimation_delay=4.0,
            decimation_correction=True)) == \
            "Response type: PolynomialResponseStage, " \
            "Stage Sequence Number: 2\n" \
            "\tFrom PA (Pascal) to COUNTS (digital_counts)\n" \
            "\tStage gain: 12345.0, defined at 1.00 Hz\n" \
            "\tDecimation:\n" \
            "\t\tInput Sample Rate: 1.00 Hz\n" \
            "\t\tDecimation Factor: 2\n" \
            "\t\tDecimation Offset: 3\n" \
            "\t\tDecimation Delay: 4.00\n" \
            "\t\tDecimation Correction: 1.00\n" \
            "\tPolynomial approximation type: MACLAURIN\n" \
            "\tFrequency lower bound: 1.0\n" \
            "\tFrequency upper bound: 2.0\n" \
            "\tApproximation lower bound: 3.0\n" \
            "\tApproximation upper bound: 4.0\n" \
            "\tMaximum error: 1.5\n" \
            "\tNumber of coefficients: 3"

        # Now only the very minimum.
        assert str(PolynomialResponseStage(
            stage_sequence_number=4,
            stage_gain=None,
            stage_gain_frequency=None,
            input_units=None,
            input_units_description=None,
            output_units=None,
            output_units_description=None,
            frequency_lower_bound=None,
            frequency_upper_bound=None,
            approximation_lower_bound=None,
            approximation_upper_bound=None,
            maximum_error=None,
            coefficients=[],
            approximation_type="MACLAURIN")) == \
            "Response type: PolynomialResponseStage, " \
            "Stage Sequence Number: 4\n" \
            "\tFrom UNKNOWN to UNKNOWN\n" \
            "\tStage gain: UNKNOWN, defined at UNKNOWN Hz\n" \
            "\tPolynomial approximation type: MACLAURIN\n" \
            "\tFrequency lower bound: None\n" \
            "\tFrequency upper bound: None\n" \
            "\tApproximation lower bound: None\n" \
            "\tApproximation upper bound: None\n" \
            "\tMaximum error: None\n" \
            "\tNumber of coefficients: 0"

    def test_get_sampling_rates(self, testdata):
        """
        Tests for the get_sampling_rates() method.
        """
        # Test for the default inventory.
        resp = read_inventory()[0][0][0].response
        assert resp.get_sampling_rates() == \
            {1: {'decimation_factor': 1,
                 'input_sampling_rate': 200.0,
                 'output_sampling_rate': 200.0},
             2: {'decimation_factor': 1,
                 'input_sampling_rate': 200.0,
                 'output_sampling_rate': 200.0}}

        # Another, well behaved file.
        inv = read_inventory(testdata['AU.MEEK.xml'])
        assert inv[0][0][0].response.get_sampling_rates() == \
            {1: {'decimation_factor': 1,
                 'input_sampling_rate': 600.0,
                 'output_sampling_rate': 600.0},
             2: {'decimation_factor': 1,
                 'input_sampling_rate': 600.0,
                 'output_sampling_rate': 600.0},
             3: {'decimation_factor': 1,
                 'input_sampling_rate': 600.0,
                 'output_sampling_rate': 600.0},
             4: {'decimation_factor': 3,
                 'input_sampling_rate': 600.0,
                 'output_sampling_rate': 200.0},
             5: {'decimation_factor': 10,
                 'input_sampling_rate': 200.0,
                 'output_sampling_rate': 20.0}}

        # This file lacks decimation attributes for the first two stages as
        # well as one of the later ones. These thus have to be inferred.
        inv = read_inventory(testdata['DK.BSD..BHZ.xml'])
        assert inv[0][0][0].response.get_sampling_rates() == \
            {1: {'decimation_factor': 1,
                 'input_sampling_rate': 30000.0,
                 'output_sampling_rate': 30000.0},
             2: {'decimation_factor': 1,
                 'input_sampling_rate': 30000.0,
                 'output_sampling_rate': 30000.0},
             3: {'decimation_factor': 1,
                 'input_sampling_rate': 30000.0,
                 'output_sampling_rate': 30000.0},
             4: {'decimation_factor': 5,
                 'input_sampling_rate': 30000.0,
                 'output_sampling_rate': 6000.0},
             5: {'decimation_factor': 3,
                 'input_sampling_rate': 6000.0,
                 'output_sampling_rate': 2000.0},
             6: {'decimation_factor': 2,
                 'input_sampling_rate': 2000.0,
                 'output_sampling_rate': 1000.0},
             7: {'decimation_factor': 5,
                 'input_sampling_rate': 1000.0,
                 'output_sampling_rate': 200.0},
             8: {'decimation_factor': 2,
                 'input_sampling_rate': 200.0,
                 'output_sampling_rate': 100.0},
             9: {'decimation_factor': 1,
                 'input_sampling_rate': 100.0,
                 'output_sampling_rate': 100.0},
             10: {'decimation_factor': 5,
                  'input_sampling_rate': 100.0,
                  'output_sampling_rate': 20.0}}

    def test_response_calculation_paz_without_decimation(self, testdata):
        """
        This test files has two PAZ stages with no decimation attributes.

        Evalresp does not like this so we have to add dummy decimation
        attributes before calling it.
        """
        inv = read_inventory(testdata['DK.BSD..BHZ.xml'])
        np.testing.assert_allclose(
            inv[0][0][0].response.get_evalresp_response_for_frequencies(
                [0.1, 1.0, 10.0], hide_sensitivity_mismatch_warning=True),
            [6.27191825e+08 + 1.38925202e+08j,
             6.51826202e+08 + 1.28404787e+07j,
             2.00067263e+04 - 2.63711751e+03j])

    def test_regression_evalresp(self, testdata):
        """
        Regression test for an evalresp issue with a micropressure instrument.

        See #2171.
        """
        inv = read_inventory(testdata["IM_I53H1_BDF.xml"])
        cha_response = inv[0][0][0].response
        freq_resp = cha_response.get_evalresp_response_for_frequencies([0.0])
        assert freq_resp == 0.0 + 0.0j
        np.testing.assert_allclose(
            inv[0][0][0].response.get_evalresp_response_for_frequencies(
                [0.1, 1.0, 10.0]),
            [2.411908e+05 + 2.283852e+04j,
             2.445572e+05 - 2.480459e+03j,
             -2.455459e-01 + 4.888214e-02j], rtol=1e-6)

    def test_recalculate_overall_sensitivity(self):
        """
        Tests the recalculate_overall_sensitivity_method().

        This is not yet an exhaustive test as responses are complicated...
        """
        resp = read_inventory()[0][0][0].response
        np.testing.assert_allclose(
            resp.instrument_sensitivity.value,
            943680000.0)
        np.testing.assert_allclose(
            resp.instrument_sensitivity.frequency,
            0.02)
        # Recompute - it is not much different but a bit.
        resp.recalculate_overall_sensitivity(0.02)
        np.testing.assert_allclose(
            resp.instrument_sensitivity.value,
            943681500.0)
        np.testing.assert_allclose(
            resp.instrument_sensitivity.frequency,
            0.02)

        # There is some logic to automatically determine a suitable frequency.
        # Make sure this does something here.
        resp = read_inventory()[0][0][0].response
        resp.recalculate_overall_sensitivity()
        np.testing.assert_allclose(
            resp.instrument_sensitivity.value,
            957562105.3939067)
        np.testing.assert_allclose(
            resp.instrument_sensitivity.frequency,
            1.0)

        # Passing an integer also works. See #2338.
        resp = read_inventory()[0][0][0].response
        resp.recalculate_overall_sensitivity(1)
        np.testing.assert_allclose(
            resp.instrument_sensitivity.value,
            957562105.3939067)
        np.testing.assert_allclose(
            resp.instrument_sensitivity.frequency,
            1.0)

    def test_warning_response_list_extrapolation(self):
        """
        Tests that a warning message is shown when a response calculation
        involving a response list stage includes (spline) extrapolation into
        frequency ranges outside of the band defined by the response list stage
        elements (frequency-amplitude-phase pairs).
        """
        # create some bogus response list stage for the test
        elems = [
            ResponseListElement(x, 1, 0) for x in (1, 2, 5, 10, 20, 40, 50)]
        stage = ResponseListResponseStage(
            1, 1, 10, "M/S", "M/S", response_list_elements=elems)
        sens = InstrumentSensitivity(1, 10, "M/S", "M/S")
        resp = Response(response_stages=[stage], instrument_sensitivity=sens)
        msg = ("The response contains a response list stage with frequencies "
               "only from 1.0000 - 50.0000 Hz. You are requesting a response "
               "from 25.0000 - 100.0000 Hz. The calculated response will "
               "contain extrapolation beyond the frequency band constrained "
               "by the response list stage. Please consider adjusting "
               "'pre_filt' and/or 'water_level' during response removal "
               "accordingly.")
        with CatchAndAssertWarnings(expected=[(UserWarning, msg)]):
            resp.get_evalresp_response(0.005, 2**3)

    def test_unknown_units_no_integration(self, testdata):
        """
        Makes sure that when a unit in the list of response stages is not known
        to evalresp, no tampering (integration/differentiation) is done and
        response is just calculated as is specified.
        It used to be that unkown units where reported to evalresp as
        displacement ("DIS") which led to a differentiation being added
        internally with the default "output='VEL'".
        Example is a rotational sensor with input units RAD/S and a flat
        response and since evalresp does not know RAD or RAD/S the only thing
        that makes sense is have evalresp not do anything else than use the
        response as is and ignore "output" option.
        """
        inv = read_inventory(testdata['response_radian_per_second.xml'],
                             format="STATIONXML")
        resp = inv[0][0][0].response
        freqs = [0.01, 0.1, 1, 10, 100]
        overall_sensitivity = 1.00000000e+09
        msg = (r"The unit 'RAD/S' is not known to ObsPy. It will be passed in "
               r"to evalresp as 'undefined'. This should result in evalresp "
               r"using the response as is, without adding any integration or "
               r"differentiation and the 'output' parameter \(here: 'VEL'\) "
               r"not having any effect. Please double check output data.")
        with pytest.warns(UserWarning, match=msg):
            data = resp.get_evalresp_response_for_frequencies(freqs)
        assert resp.instrument_sensitivity.value == overall_sensitivity
        assert np.abs(data).tolist() == [overall_sensitivity] * len(freqs)

    def test_non_SI_unit_input_first_stage(self, testdata):
        """
        Regression test for #3369

        Test that non-SI units input (e.g. nanometer/second) is properly taken
        into account when removing the instrument response, i.e. the additional
        scaling factor (e.g. 1e09 for nm/s) is applied to the output.

        The change in that issue affects
        Response._call_eval_resp_for_frequencies() which in turn gets called
        by..
         - Response._get_overall_sensitivity_and_gain()
         - Response.get_evalresp_response_for_frequencies()

        Test all of those three for good measure.
        """
        inv = read_inventory(testdata['SL_BOJS_LHZ.xml'], "STATIONXML")

        # original data has nm/s as a non-SI input unit. setup clones with some
        # different input units
        inv_nms = inv
        inv_ms = inv_nms.copy()
        inv_ms[0][0][0].response.instrument_sensitivity.input_units = 'm/s'
        inv_ms[0][0][0].response.instrument_sensitivity.value *= 1e9
        inv_ms[0][0][0].response.response_stages[0].input_units = 'm/s'
        inv_ms[0][0][0].response.response_stages[0].stage_gain *= 1e9
        inv_mms = inv_nms.copy()
        inv_mms[0][0][0].response.instrument_sensitivity.input_units = 'mm/s'
        inv_mms[0][0][0].response.instrument_sensitivity.value *= 1e6
        inv_mms[0][0][0].response.response_stages[0].input_units = 'mm/s'
        inv_mms[0][0][0].response.response_stages[0].stage_gain *= 1e6
        inv_cms = inv_nms.copy()
        inv_cms[0][0][0].response.instrument_sensitivity.input_units = 'cm/s'
        inv_cms[0][0][0].response.instrument_sensitivity.value *= 1e7
        inv_cms[0][0][0].response.response_stages[0].input_units = 'cm/s'
        inv_cms[0][0][0].response.response_stages[0].stage_gain *= 1e7

        # test at two frequencies in the flat response part
        freqs = [0.01, 0.1]
        expected = [1.84e+09, 1.85e+09]

        for inv in (inv_nms, inv_ms, inv_mms, inv_cms):
            resp = inv[0][0][0].response

            # test _call_eval_resp_for_frequencies()
            got, _ = resp._call_eval_resp_for_frequencies(
                freqs, output='VEL', start_stage=None, end_stage=None)
            got = np.abs(got)
            np.testing.assert_allclose(got, expected, rtol=1e-2)

            # test get_evalresp_response_for_frequencies()
            got = resp.get_evalresp_response_for_frequencies(
                freqs, output='VEL', start_stage=None, end_stage=None)
            got = np.abs(got)
            np.testing.assert_allclose(got, expected, rtol=1e-2)

            # test _get_overall_sensitivity_and_gain()
            got = [resp._get_overall_sensitivity_and_gain(f, output='VEL')[1]
                   for f in freqs]
            np.testing.assert_allclose(got, expected, rtol=1e-2)

    def test_pazresponsestage_hertz_to_radians(self, testdata):
        """
        Tests converting a PAZResponseStage from Hertz to Radians/s
        """
        inv = read_inventory(testdata['G_CAN__LHZ.xml'], 'STATIONXML')
        paz_before = inv[0][0][0].response.response_stages[0]
        paz_after = deepcopy(paz_before)
        paz_after.to_radians_per_second()

        expected_zeros = [0j, 0j]
        expected_poles = [-1.233948e-02+1.234319e-02j,
                          -1.233948e-02-1.234319e-02j,
                          -3.917566e+01+4.912339e+01j,
                          -3.917566e+01-4.912339e+01j]

        assert paz_before.pz_transfer_function_type == 'LAPLACE (HERTZ)'
        assert paz_after.pz_transfer_function_type == \
            'LAPLACE (RADIANS/SECOND)'
        assert round(paz_after.normalization_factor, 3) == 3.959488e+03
        np.testing.assert_allclose(
            paz_after.zeros, expected_zeros, rtol=1e-6)
        np.testing.assert_allclose(
            paz_after.poles, expected_poles, rtol=1e-6)

    def test_unknown_units_PA_recalculate_sensitivity(self, testdata):
        """
        Test recalculating overall sensitivity in presence of unusual units, in
        this case Pascal.
        """
        inv = read_inventory(testdata['hydrophone_response_PA.xml'],
                             format="STATIONXML")
        resp = inv[0][0][0].response
        msg = ("ObsPy can not map unit 'PA' to "
               "displacement, velocity, or acceleration - "
               "evalresp should still work and just use the response as "
               "is. This might not be covered by tests, though, so "
               "proceed with caution and report any unexpected "
               "behavior.")
        with pytest.warns(UserWarning, match=msg):
            resp.recalculate_overall_sensitivity()
        assert np.isclose(
            resp.instrument_sensitivity.value, 133579131859239.3, atol=0,
            rtol=1e-5)

    def test_polynomial_response(self, testdata):
        """
        Test interpretation of 2nd order polynomial response
        (modified from NZ.CHIT.41.LTZ)
        """
        inv = read_inventory(testdata['polynomial_response.xml'],
                             format="STATIONXML")
        resp = inv[0][0][0].response
        msg = (r"PolynomialResponseStage \(stage 1\) has a DC "
               "offset of -4.854369 which is ignored in "
               "frequency domain calculations.")
        with pytest.warns(UserWarning, match=msg):
            resp.recalculate_overall_sensitivity()
        assert np.isclose(
            resp.instrument_sensitivity.value, 82400.0, atol=0,
            rtol=1e-5)



# -*- coding: utf-8 -*-
import os
import copy
from unittest import mock

import numpy as np
import pytest
from requests import HTTPError

from obspy.core.util.base import (NamedTemporaryFile, get_dependency_version,
                                  download_to_file, sanitize_filename,
                                  create_empty_data_chunk, ComparingObject)


class TestUtilBase:
    """
    Test suite for obspy.core.util.base
    """
    def test_get_matplotlib_version(self):
        """
        Tests for the get_matplotlib_version() function as it continues to
        cause problems.
        """
        versions = (("1.2.3", [1, 2, 3]), ("0.9.11", [0, 9, 11]),
                    ("0.9.svn", [0, 9, 0]), ("1.1.1~rc1-1", [1, 1, 1]),
                    ("1.2.x", [1, 2, 0]), ("1.3.1rc2", [1, 3, 1]))

        for version_string, expected in versions:
            with mock.patch('importlib.metadata.version') as p:
                p.return_value = version_string
                got = get_dependency_version('matplotlib')
            assert expected == got

    def test_named_temporay_file__context_manager(self):
        """
        Tests the automatic closing/deleting of NamedTemporaryFile using the
        context manager.
        """
        content = b"burn after writing"
        # write something to tempfile and check closing/deletion afterwards
        with NamedTemporaryFile() as tf:
            filename = tf.name
            tf.write(content)
        assert not os.path.exists(filename)
        # write something to tempfile and check that it is written correctly
        with NamedTemporaryFile() as tf:
            filename = tf.name
            tf.write(content)
            tf.close()
            with open(filename, 'rb') as fh:
                tmp_content = fh.read()
        assert content == tmp_content
        assert not os.path.exists(filename)
        # check that closing/deletion works even when nothing is done with file
        with NamedTemporaryFile() as tf:
            filename = tf.name
        assert not os.path.exists(filename)

    def test_mock_read_inventory_http_errors(self):
        """
        Tests HTTP Error on 204, 400, and 500
        """
        url = "http://obspy.org"
        for response_tuple in [("204", "No Content"), ("400", "Bad Request"),
                               ("500", "Internal Server Error")]:
            code = response_tuple[0]
            reason = response_tuple[1]
            with mock.patch("requests.get") as mocked_get:
                mocked_get.return_value.status_code = code
                mocked_get.return_value.reason = reason
                msg = "%s HTTP Error: %s for url: %s" % (code, reason, url)
                with pytest.raises(HTTPError, match=msg):
                    download_to_file(url, None)

    def test_sanitize_filename(self):
        assert sanitize_filename("example.mseed") == \
               "example.mseed"
        assert sanitize_filename("Example.mseed") == \
               "Example.mseed"
        assert sanitize_filename("example.mseed?raw=True") == \
               "example.mseedrawTrue"
        assert sanitize_filename("Example.mseed?raw=true") == \
               "Example.mseedrawtrue"

    def test_create_empty_data_chunk(self):
        out = create_empty_data_chunk(3, 'int', 10)
        assert isinstance(out, np.ndarray)
        # The default dtype for an integer (np.int_) is a `C long` which is
        # only 32 bits on windows. Thus we have to allow both.
        assert out.dtype in (np.int32, np.int64)
        np.testing.assert_allclose(out, [10, 10, 10])

        out = create_empty_data_chunk(6, np.complex128, 0)
        assert isinstance(out, np.ndarray)
        assert out.dtype == np.complex128
        np.testing.assert_allclose(out, np.zeros(6, dtype=np.complex128))

        # Fully masked output.
        out = create_empty_data_chunk(3, 'f')
        assert isinstance(out, np.ma.MaskedArray)
        assert out.dtype == np.float32
        np.testing.assert_allclose(out.mask, [True, True, True])

    def test_comparing_object_eq(self):
        co = ComparingObject()
        # Compare to other types
        assert co != 5
        assert co is not None
        assert co != object()
        # Compare same type, different instance, with attributes
        co.at = 3
        deep_copy = copy.deepcopy(co)
        assert co == deep_copy
        deep_copy.at = 0
        assert co != deep_copy



import obspy
from obspy.signal.trigger import plot_trigger, z_detect


trace = obspy.read("https://examples.obspy.org/ev0_6.a01.gse2")[0]
df = trace.stats.sampling_rate

cft = z_detect(trace.data, int(10. * df))
plot_trigger(trace, cft, -0.4, -0.3)



# -*- coding: utf-8 -*-
"""
obspy.clients.iris - EarthScope (former IRIS) web service client for ObsPy
==========================================================================
The obspy.clients.iris package contains a client for the DMC Web services
provided by EarthScope (former IRIS) (https://service.earthscope.org/irisws/).

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Local Alternatives
------------------

ObsPy contains local alternatives to some of the calculation tools offered by
the EarthScope web services. Consider using them when working within ObsPy:

+---------------------------------------------------------+--------------------------------------------------------------+
| EarthScope Web Service                                  | Equivalent ObsPy Function/Module                             |
+=========================================================+==============================================================+
| :meth:`obspy.clients.iris.client.Client.traveltime()`   | :mod:`obspy.taup`                                            |
+---------------------------------------------------------+--------------------------------------------------------------+
| :meth:`obspy.clients.iris.client.Client.distaz()`       | :mod:`obspy.geodetics`                                       |
+---------------------------------------------------------+--------------------------------------------------------------+
| :meth:`obspy.clients.iris.client.Client.flinnengdahl()` | :class:`obspy.geodetics.flinnengdahl.FlinnEngdahl`           |
+---------------------------------------------------------+--------------------------------------------------------------+

Web service Interfaces
----------------------

Each of the following methods directly wrap a single Web service provided by
EarthScope (https://service.earthscope.org/irisws/):

**Request Tools**

* :meth:`~obspy.clients.iris.client.Client.evalresp()` - evaluates instrument
  response information stored at the EarthScope DMC and outputs ASCII data or
  `Bode Plots <https://en.wikipedia.org/wiki/Bode_plots>`_.
* :meth:`~obspy.clients.iris.client.Client.resp()` - provides access to channel
  response information in the SEED RESP format (as used by evalresp)
* :meth:`~obspy.clients.iris.client.Client.sacpz()` - provides access to
  instrument response information (per-channel) as poles and zeros in the ASCII
  format used by SAC and other programs
* :meth:`~obspy.clients.iris.client.Client.timeseries()` - fetches segments of
  seismic data and returns data formatted in either MiniSEED, ASCII or SAC. It
  can optionally filter the data.

**Calculation Tools**

* :meth:`~obspy.clients.iris.client.Client.traveltime()` - calculates
  travel-times for seismic phases using a 1-D spherical Earth model.
* :meth:`~obspy.clients.iris.client.Client.distaz()` - calculate the distance
  and azimuth between two points on a sphere.
* :meth:`~obspy.clients.iris.client.Client.flinnengdahl()` - converts a
  latitude, longitude pair into either a Flinn-Engdahl seismic region code or
  region name.


Please see the documentation for each method for further information and
examples to retrieve various data from the EarthScope DMC.
"""
from .client import Client  # NOQA


__all__ = ["Client"]


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
obspy.clients.seedlink - SeedLink client for ObsPy
==================================================

The obspy.clients.seedlink module provides an implementation of the SeedLink
client protocol for ObsPy.

For simple requests of finite time windows see
:class:`~obspy.clients.seedlink.basic_client.Client`. To work with continuous
data streams see
:class:`~obspy.clients.seedlink.easyseedlink.EasySeedLinkClient`, or for
lower-level packet handling see
:class:`~obspy.clients.seedlink.slclient.SLClient`.

:copyright:
    The ObsPy Development Team (devs@obspy.org) & Anthony Lomax
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

The obspy.clients.seedlink development has been supported by the NERA project
["Network of European Research Infrastructures for Earthquake Risk Assessment
and Mitigation" under the European Community's Seventh Framework Programme
(FP7/2007-2013) grant agreement n° 262330] and implemented within the
activities of the JRA2/WP12 "Tools for real-time seismology, acquisition and
mining".
"""
from .basic_client import Client  # NOQA
from .slclient import SLClient  # NOQA
from .easyseedlink import EasySeedLinkClient  # NOQA

if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



from obspy import read_events, read_inventory
from obspy.io.sh.evt import _is_evt


class TestEvt():
    def test_is_evt_file(self, datapath):
        for fname in datapath.glob('*.evt'):
            assert _is_evt(fname)

    def test_local_event1(self, testdata):
        fname = testdata['local1.evt']
        cat = read_events(fname)
        assert len(cat) == 2

    def test_local_event2(self, testdata):
        fname = testdata['local2.evt']
        cat = read_events(fname)
        assert len(cat) == 1
        ev = cat[0]
        mag = ev.preferred_magnitude()
        origin = ev.preferred_origin()
        oc = origin.origin_uncertainty
        assert mag.mag == 0.6
        assert mag.magnitude_type == 'ML'
        assert round(origin.longitude_errors.uncertainty, 3) == 0.013
        assert round(origin.latitude_errors.uncertainty, 3) == 0.013
        assert oc.min_horizontal_uncertainty == 20
        assert oc.max_horizontal_uncertainty == 20
        assert oc.azimuth_max_horizontal_uncertainty == 75.70
        assert origin.arrivals[0].time_weight == 4.0

    def test_tele_event1(self, testdata):
        fname = testdata['tele1.evt']
        cat = read_events(fname)
        assert len(cat) == 1

    def test_tele_event2(self, testdata):
        # untested field: sign
        fname = testdata['tele2.evt']
        cat = read_events(fname)
        assert len(cat) == 1
        ev = cat[0]
        origin = ev.preferred_origin()
        mag = ev.preferred_magnitude()
        sta_mag = ev.station_magnitudes[0]
        pick = ev.picks[66]
        arrival = origin.arrivals[66]
        arrival = origin.arrivals[191]
        assert len(ev.picks) == 195
        assert len(ev.station_magnitudes) == 38
        assert len(origin.arrivals) == 195

        assert str(pick.time) == '2015-08-10T10:14:32.918000Z'
        assert pick.phase_hint == 'T'
        assert pick.onset == 'emergent'
        assert pick.evaluation_mode == 'manual'
        assert str(pick.filter_id) == 'G_WWSSN_SP'
        assert pick.waveform_id.get_seed_string() == '.GRA1..Z'
        assert origin.arrivals[66].pick_id == pick.resource_id
        assert arrival.phase == 'sS'
        assert arrival.azimuth == 307.6
        assert arrival.distance == 48.132
        assert str(origin.time) == '2015-08-10T10:05:25.808000Z'
        assert origin.latitude == 36.23
        assert origin.longitude == 71.38

        assert origin.depth == 238200.0
        assert origin.depth_errors.uncertainty == 7180.0
        assert origin.quality.used_station_count == 30
        assert origin.region == 'Afghanistan-Tajikistan border region'
        assert ev.event_type == 'earthquake'

        assert mag.mag == 6.1
        assert mag.magnitude_type == 'Mb'
        assert sta_mag.mag == 6.2
        assert sta_mag.station_magnitude_type == 'Mb'
        assert sta_mag.waveform_id.get_seed_string() == '.AHRW..Z'

    def test_populate_waveform_id(self, testdata):
        fname = testdata['tele2.evt']
        # read invenroty - FUR with channels, WET without channels
        inv = read_inventory().select(channel='HH?')
        inv[0][1].channels = []
        cat = read_events(fname, inventory=inv,
                          default_seedid='BLA.{}.11.DH{}',
                          seedid_map={'UBR': 'BLB.{}.00.BH{}'}, warn=False)
        picks = cat[0].picks
        waveform_ids = [p.waveform_id.get_seed_string() for p in picks]
        assert 'GR.FUR..HHZ' in waveform_ids
        assert 'BLA.WET.11.DHZ' in waveform_ids
        assert 'BLB.UBR.00.BHZ' in waveform_ids
        assert 'BLA.WERD.11.DHZ' in waveform_ids



"""Implements (a subset of) Sun XDR -- eXternal Data Representation.
See: RFC 1014

PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
--------------------------------------------

1. This LICENSE AGREEMENT is between the Python Software Foundation
("PSF"), and the Individual or Organization ("Licensee") accessing and
otherwise using this software ("Python") in source or binary form and
its associated documentation.

2. Subject to the terms and conditions of this License Agreement, PSF hereby
grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce,
analyze, test, perform and/or display publicly, prepare derivative works,
distribute, and otherwise use Python alone or in any derivative version,
provided, however, that PSF's License Agreement and PSF's notice of copyright,
i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
2010,2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022
Python Software Foundation;
All Rights Reserved" are retained in Python alone or in any derivative version
prepared by Licensee.

3. In the event Licensee prepares a derivative work that is based on
or incorporates Python or any part thereof, and wants to make
the derivative work available to others as provided herein, then
Licensee hereby agrees to include in any such work a brief summary of
the changes made to Python.

4. PSF is making Python available to Licensee on an "AS IS"
basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND
DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT
INFRINGE ANY THIRD PARTY RIGHTS.

5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS
A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON,
OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material
breach of its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any
relationship of agency, partnership, or joint venture between PSF and
Licensee.  This License Agreement does not grant permission to use PSF
trademarks or trade name in a trademark sense to endorse or promote
products or services of Licensee, or any third party.

8. By copying, installing or otherwise using Python, Licensee
agrees to be bound by the terms and conditions of this License
Agreement.
"""

import struct
from io import BytesIO
from functools import wraps


__all__ = ["Error", "Packer", "Unpacker", "ConversionError"]

# exceptions


class Error(Exception):
    """Exception class for this module. Use:
    except xdrlib.Error as var:
        # var has the Error instance for the exception
    Public ivars:
        msg -- contains the message
    """

    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return repr(self.msg)

    def __str__(self):
        return str(self.msg)


class ConversionError(Error):
    pass


def raise_conversion_error(function):
    """ Wrap any raised struct.errors in a ConversionError. """

    @wraps(function)
    def result(self, value):
        try:
            return function(self, value)
        except struct.error as e:
            raise ConversionError(e.args[0]) from None
    return result


class Packer:
    """Pack various data representations into a buffer."""

    def __init__(self):
        self.reset()

    def reset(self):
        self.__buf = BytesIO()

    def get_buffer(self):
        return self.__buf.getvalue()
    # backwards compatibility
    get_buf = get_buffer

    @raise_conversion_error
    def pack_uint(self, x):
        self.__buf.write(struct.pack('>L', x))

    @raise_conversion_error
    def pack_int(self, x):
        self.__buf.write(struct.pack('>l', x))

    pack_enum = pack_int

    def pack_bool(self, x):
        if x:
            self.__buf.write(b'\0\0\0\1')
        else:
            self.__buf.write(b'\0\0\0\0')

    def pack_uhyper(self, x):
        try:
            self.pack_uint(x >> 32 & 0xffffffff)
        except (TypeError, struct.error) as e:
            raise ConversionError(e.args[0]) from None
        try:
            self.pack_uint(x & 0xffffffff)
        except (TypeError, struct.error) as e:
            raise ConversionError(e.args[0]) from None

    pack_hyper = pack_uhyper

    @raise_conversion_error
    def pack_float(self, x):
        self.__buf.write(struct.pack('>f', x))

    @raise_conversion_error
    def pack_double(self, x):
        self.__buf.write(struct.pack('>d', x))

    def pack_fstring(self, n, s):
        if n < 0:
            raise ValueError('fstring size must be nonnegative')
        data = s[:n]
        n = ((n+3)//4)*4
        data = data + (n - len(data)) * b'\0'
        self.__buf.write(data)

    pack_fopaque = pack_fstring

    def pack_string(self, s):
        n = len(s)
        self.pack_uint(n)
        self.pack_fstring(n, s)

    pack_opaque = pack_string
    pack_bytes = pack_string

    def pack_list(self, list, pack_item):
        for item in list:
            self.pack_uint(1)
            pack_item(item)
        self.pack_uint(0)

    def pack_farray(self, n, list, pack_item):
        if len(list) != n:
            raise ValueError('wrong array size')
        for item in list:
            pack_item(item)

    def pack_array(self, list, pack_item):
        n = len(list)
        self.pack_uint(n)
        self.pack_farray(n, list, pack_item)


class Unpacker:
    """Unpacks various data representations from the given buffer."""

    def __init__(self, data):
        self.reset(data)

    def reset(self, data):
        self.__buf = data
        self.__pos = 0

    def get_position(self):
        return self.__pos

    def set_position(self, position):
        self.__pos = position

    def get_buffer(self):
        return self.__buf

    def done(self):
        if self.__pos < len(self.__buf):
            raise Error('unextracted data remains')

    def unpack_uint(self):
        i = self.__pos
        self.__pos = j = i+4
        data = self.__buf[i:j]
        if len(data) < 4:
            raise EOFError
        return struct.unpack('>L', data)[0]

    def unpack_int(self):
        i = self.__pos
        self.__pos = j = i+4
        data = self.__buf[i:j]
        if len(data) < 4:
            raise EOFError
        return struct.unpack('>l', data)[0]

    unpack_enum = unpack_int

    def unpack_bool(self):
        return bool(self.unpack_int())

    def unpack_uhyper(self):
        hi = self.unpack_uint()
        lo = self.unpack_uint()
        return int(hi) << 32 | lo

    def unpack_hyper(self):
        x = self.unpack_uhyper()
        if x >= 0x8000000000000000:
            x = x - 0x10000000000000000
        return x

    def unpack_float(self):
        i = self.__pos
        self.__pos = j = i+4
        data = self.__buf[i:j]
        if len(data) < 4:
            raise EOFError
        return struct.unpack('>f', data)[0]

    def unpack_double(self):
        i = self.__pos
        self.__pos = j = i+8
        data = self.__buf[i:j]
        if len(data) < 8:
            raise EOFError
        return struct.unpack('>d', data)[0]

    def unpack_fstring(self, n):
        if n < 0:
            raise ValueError('fstring size must be nonnegative')
        i = self.__pos
        j = i + (n+3)//4*4
        if j > len(self.__buf):
            raise EOFError
        self.__pos = j
        return self.__buf[i:i+n]

    unpack_fopaque = unpack_fstring

    def unpack_string(self):
        n = self.unpack_uint()
        return self.unpack_fstring(n)

    unpack_opaque = unpack_string
    unpack_bytes = unpack_string

    def unpack_list(self, unpack_item):
        list = []
        while 1:
            x = self.unpack_uint()
            if x == 0:
                break
            if x != 1:
                raise ConversionError('0 or 1 expected, got %r' % (x,))
            item = unpack_item()
            list.append(item)
        return list

    def unpack_farray(self, n, unpack_item):
        list = []
        for i in range(n):
            list.append(unpack_item())
        return list

    def unpack_array(self, unpack_item):
        n = self.unpack_uint()
        return self.unpack_farray(n, unpack_item)



# -*- coding: utf-8 -*-
"""
Support for reading CyberShake seismograms in ObsPy.
"""
import io
import struct
from pathlib import Path

import numpy as np

from obspy import Trace, Stream


NETWORK_CODE = 'CS'  # Assign a network code for CyberShake data
INSTRUMENT_CODE = 'X'  # X indicates synthetic data
LOCATION_CODE = '00'


def _is_cybershake(filename):
    """
    Checks whether a file is a CyberShake version 12.10 seismogram file or not.

    :param filename: file to be checked.
    :type filename: str, open file, or file-like object
    :rtype: bool
    :return: ``True`` if a CyberShake version 12.0 seismogram file.
    """

    if isinstance(filename, io.BufferedIOBase):
        starting_pos = filename.tell()
        is_cybershake_result = _internal_is_cybershake(filename)
        filename.seek(starting_pos, 0)
        return is_cybershake_result
    elif isinstance(filename, (str, bytes, Path)):
        with open(filename, "rb") as fh:
            return _internal_is_cybershake(fh)
    else:
        raise ValueError("Cannot open '%s'." % filename)


def _internal_is_cybershake(buf):
    """
    Checks whether a file-like object contains a CyberShake file or not.

    :param buf: CyberShake file to be checked.
    :type buf: file-like object or open file
    :rtype: bool
    :return: ``True`` if a CyberShake file.
    """
    try:
        header_str = buf.read(56)
        split_point = header_str[0:8].find(b'\x00')
        version = header_str[0:split_point].decode()
        assert version == "12.10"
        return True
    except Exception:
        return False


def _read_cybershake(filename, band_code='M', **kwargs):
    """
    Reads a CyberShake seismogram file and returns a Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param filename: Filename that contains the CyberShake seismogram data.
    :param band_code: A single character associated with the period band.
        Default is 'M' for mid-period.
    """
    if isinstance(filename, io.BufferedIOBase):
        return _internal_read_cybershake(buf=filename, band_code=band_code,
                                         **kwargs)
    else:
        with open(filename, "rb") as fh:
            return _internal_read_cybershake(buf=fh, band_code=band_code,
                                             **kwargs)


def _internal_read_cybershake(buf, band_code='M', **kwargs):
    """
    Reads a CyberShake seismogram file and returns a Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :param buf: Filename that contains the CyberShake seismogram data.
    :type buf: file or file-like object
    :param band_code: A single character associated with the period band.
        Default is 'M' for mid-period.
    :type band_code: str
    """

    header_str = buf.read(56)
    split_point = header_str[8:16].find(b'\x00')
    site = header_str[8:8+split_point].decode()

    source_id = struct.unpack('i', header_str[24:28])[0]
    rupture_id = struct.unpack('i', header_str[28:32])[0]
    rup_var_id = struct.unpack('i', header_str[32:36])[0]
    dt = struct.unpack('f', header_str[36:40])[0]
    nt = struct.unpack('i', header_str[40:44])[0]
    det_max_freq = struct.unpack('f', header_str[48:52])[0]
    stoch_max_freq = struct.unpack('f', header_str[52:56])[0]

    data_str = buf.read(4 * nt)
    x_data = np.array(struct.unpack("%df" % nt, data_str))
    data_str = buf.read(4 * nt)
    y_data = np.array(struct.unpack("%df" % nt, data_str))

    traces = []
    for data, orientation_code in zip([x_data, y_data], ['E', 'N']):
        header = {
            'network': NETWORK_CODE,
            'station': site,
            'location': LOCATION_CODE,
            'channel': band_code + INSTRUMENT_CODE + orientation_code,
            'delta': dt,
            'cybershake': {
                'source_id': source_id,
                'rupture_id': rupture_id,
                'rup_var_id': rup_var_id,
                'det_max_freq': det_max_freq,
                'stoch_max_freq': stoch_max_freq
            }
        }
        traces.append(Trace(data, header))
    return Stream(traces)



# -*- coding: utf-8 -*-
"""
GSE2.0 bulletin read support.

:author:
    EOST (École et Observatoire des Sciences de la Terre)
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import re
import warnings

from obspy.core.event import (Amplitude, Arrival, Catalog, Comment,
                              CreationInfo, Event, EventDescription,
                              Magnitude, Origin, OriginQuality,
                              OriginUncertainty, Pick, ResourceIdentifier,
                              StationMagnitude,
                              StationMagnitudeContribution,
                              WaveformStreamID)
from obspy.core.event.header import (
    EvaluationMode, EventDescriptionType, EventType, EventTypeCertainty,
    OriginDepthType, OriginUncertaintyDescription, PickOnset, PickPolarity)
from obspy.core.utcdatetime import UTCDateTime


# Convert GSE2 depth flag to ObsPy depth type
DEPTH_TYPES = {
    'f': OriginDepthType('operator assigned'),
    'd': OriginDepthType('constrained by depth phases'),
}

# Convert GSE2 analysis type to ObsPy evaluation modes
EVALUATION_MODES = {
    'm': EvaluationMode.MANUAL,
    'a': EvaluationMode.AUTOMATIC,
    'g': EvaluationMode.MANUAL,
}


# Convert GSE2 to ObsPy location methods
LOCATION_METHODS = {
    'i': 'inversion',
    'p': 'pattern recognition',
    'g': 'ground truth',
    'o': 'other',
}


# Convert GSE2 to ObsPy event types
EVENT_TYPES = {
    'ke': (EventTypeCertainty.KNOWN,
           EventType('earthquake')),
    'se': (EventTypeCertainty.SUSPECTED,
           EventType('earthquake')),
    'kr': (EventTypeCertainty.KNOWN,
           EventType('rock burst')),
    'sr': (EventTypeCertainty.SUSPECTED,
           EventType('rock burst')),
    'ki': (EventTypeCertainty.KNOWN,
           EventType('induced or triggered event')),
    'si': (EventTypeCertainty.SUSPECTED,
           EventType('induced or triggered event')),
    'km': (EventTypeCertainty.KNOWN,
           EventType('mining explosion')),
    'sm': (EventTypeCertainty.SUSPECTED,
           EventType('mining explosion')),
    'kx': (EventTypeCertainty.KNOWN,
           EventType('experimental explosion')),
    'sx': (EventTypeCertainty.SUSPECTED,
           EventType('experimental explosion')),
    'kn': (EventTypeCertainty.KNOWN,
           EventType('nuclear explosion')),
    'sn': (EventTypeCertainty.SUSPECTED,
           EventType('nuclear explosion')),
    'ls': (EventTypeCertainty.KNOWN,
           EventType('landslide')),
    'uk': (None,
           EventType('other')),
}


# Convert GSE2 to ObsPy polarity
PICK_POLARITIES = {
    'c': PickPolarity.POSITIVE,
    'd': PickPolarity.NEGATIVE,
}


# Convert GSE2 to ObsPy pick onset
PICK_ONSETS = {
    'E': PickOnset.EMERGENT,
    'I': PickOnset.IMPULSIVE,
    'Q': PickOnset.QUESTIONABLE,
}


class GSE2BulletinSyntaxError(Exception):
    """Raised when the file is not a valid GSE2 file"""


def _is_gse2(filename):
    """
    Checks whether a file is GSE2.0 format.

    :type filename: str
    :param filename: Name of the GSE2.0 file to be checked.
    :rtype: bool
    :return: ``True`` if GSE2.0 file.
    """
    try:
        with open(filename, 'rb') as fh:
            temp = fh.read(12)
    except Exception:
        return False
    if temp != b'BEGIN GSE2.0':
        return False
    return True


class LinesIterator(object):
    """
    Iterator to iterate file lines and count lines. Usefull for warning
    messages.
    """
    def __init__(self, lines):
        self.lines = iter(lines)
        self.line_nb = 0

    def __iter__(self):
        return self

    def __next__(self):
        self.line_nb += 1
        return next(self.lines)


class Unpickler(object):
    """
    De-serialize a GSE2.0 string into an ObsPy Catalog object.
    """

    def __init__(self, inventory, default_network_code, default_location_code,
                 default_channel_code, res_id_prefix, fields,
                 event_point_separator, agency):
        self.default_network_code = default_network_code
        self.default_location_code = default_location_code
        self.default_channel_code = default_channel_code
        self.inventory = inventory
        self.res_id_prefix = res_id_prefix
        self.event_point_separator = event_point_separator
        self.agency = agency
        self.author = ''

        self.fields = {
            'line_1': {
                'time': slice(0, 21),
                'time_fixf': slice(22, 23),
                'lat': slice(25, 33),
                'lon': slice(34, 43),
                'epicenter_fixf': slice(44, 45),
                'depth': slice(47, 52),
                'depth_fixf': slice(53, 54),
                'n_def': slice(56, 60),
                'n_sta': slice(61, 65),
                'gap': slice(66, 69),
                'mag_type_1': slice(71, 73),
                'mag_1': slice(73, 77),
                'mag_n_sta_1': slice(78, 80),
                'mag_type_2': slice(82, 84),
                'mag_2': slice(84, 88),
                'mag_n_sta_2': slice(89, 91),
                'mag_type_3': slice(93, 95),
                'mag_3': slice(95, 99),
                'mag_n_sta_3': slice(100, 102),
                'author': slice(104, 112),
                'id': slice(114, 122),
            },
            'line_2': {
                'rms': slice(5, 10),
                'ot_error': slice(15, 21),
                's_major': slice(25, 31),
                's_minor': slice(32, 38),
                'az': slice(40, 43),
                'depth_err': slice(49, 54),
                'min_dist': slice(56, 62),
                'max_dist': slice(63, 69),
                'mag_err_1': slice(74, 77),
                'mag_err_2': slice(85, 88),
                'mag_err_3': slice(96, 99),
                'antype': slice(104, 105),
                'loctype': slice(106, 107),
                'evtype': slice(108, 110),
            },
            'arrival': {
                'sta': slice(0, 5),
                'dist': slice(6, 12),
                'ev_az': slice(13, 18),
                'picktype': slice(19, 20),
                'direction': slice(20, 21),
                'detchar': slice(21, 22),
                'phase': slice(23, 30),
                'time': slice(31, 52),
                't_res': slice(53, 58),
                'azim': slice(59, 64),
                'az_res': slice(65, 71),
                'slow': slice(72, 77),
                's_res': slice(78, 83),
                't_def': slice(84, 85),
                'a_def': slice(85, 86),
                's_def': slice(86, 87),
                'snr': slice(88, 93),
                'amp': slice(94, 103),
                'per': slice(104, 109),
                'mag_type_1': slice(110, 112),
                'mag_1': slice(112, 116),
                'mag_type_2': slice(117, 119),
                'mag_2': slice(119, 123),
                'id': slice(124, 132),
            },
        }

        if fields:
            if 'line_1' in fields:
                self.fields['line_1'].update(fields['line_1'])
            if 'line_2' in fields:
                self.fields['line_2'].update(fields['line_2'])
            if 'arrival' in fields:
                self.fields['arrival'].update(fields['arrival'])

    def load(self, filename):
        """
        Read GSE2.0 file into ObsPy catalog object.

        :type filename: str
        :param filename: File name to read.
        :rtype: :class:`~obspy.core.event.Catalog`
        :return: ObsPy Catalog object.
        """
        with open(filename, 'r') as f:
            self.lines = LinesIterator(f.readlines())
        return self._deserialize()

    def _add_line_nb(self, message):
        """
        Add line number at the end of a str message.

        :type message: str
        :param message: Message for warnings or exceptions.
        :rtype: str
        :return: Message with line number.
        """
        return "%s, line %s" % (message, self.lines.line_nb)

    def _warn(self, message):
        """
        Display a warning message with the line number.

        :type message: str
        :param message: Message to be displayed
        """
        warnings.warn(self._add_line_nb(message))

    def _skip_empty_lines(self):
        line = next(self.lines)
        while not line or line.isspace():
            line = next(self.lines)
        return line

    def _get_res_id(self, ident, parent=None, parent_res_id=None):
        """
        Create a :class:`~obspy.core.event.resourceid.ResourceIdentifier`
        object.

        :type ident: str
        :param ident: Id of
            the :class:`~obspy.core.event.resourceid.ResourceIdentifier`.
        :type parent: :class:`~obspy.core.event.origin.Origin`,
            :class:`~obspy.core.event.event.Event` or any other object
            with a resource_id attribute.
        :param parent: The resource_id attribute of the parent will be
            used as a prefix for the new
            :class:`~obspy.core.event.resourceid.ResourceIdentifier`.
        :type parent_res_id:
            :class:`~obspy.core.event.resourceid.ResourceIdentifier`
        :param parent_res_id:
            :class:`~obspy.core.event.resourceid.ResourceIdentifier` of the
            parent.
        :rtype: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
        :return:  ResourceIdentifier object.
        """
        prefix = self.res_id_prefix
        # Put the parent id as prefix
        # Example: smi:local/origin/351412/arrival/6389611
        #          |        prefix        |     ident    |
        if parent:
            prefix = parent.resource_id.id
        elif parent_res_id:
            prefix = parent_res_id.id

        public_id = "%s/%s" % (prefix, ident)
        return ResourceIdentifier(public_id)

    def _comment(self, text):
        comment = Comment()
        comment.text = text
        comment.resource_id = ResourceIdentifier(prefix=self.res_id_prefix)
        return comment

    def _get_creation_info(self):
        creation_info = CreationInfo(creation_time=UTCDateTime())
        if self.agency:
            creation_info.agency_id = self.agency
        if self.author:
            creation_info.author = self.author
        return creation_info

    def _check_header(self, first_line):
        """
        Just check some stuff in header.

        :type first_line: str
        :param first_line: First line of header.
        """
        line_1_pattern = r'BEGIN\sGSE2.0'
        line_2_pattern = r'MSG_TYPE\s(REQUEST|DATA|SUBSCRIPTION|PROBLEM)'
        line_3_pattern = r'MSG_ID\s\w{1,20}\s?\w{1,8}'

        if not re.match(line_1_pattern, first_line):
            raise GSE2BulletinSyntaxError('Wrong GSE2.0 header')
        if not re.match(line_2_pattern, next(self.lines)):
            raise GSE2BulletinSyntaxError('Wrong message type in header')
        if not re.match(line_3_pattern, next(self.lines)):
            raise GSE2BulletinSyntaxError('Wrong message ID in header')

    def _get_channel(self, station, time):
        """
        Use inventory to retrieve channel and location code.

        :type station: str
        :param station: Station code
        :type time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param time: Only use channel active at given point in time.
        :rtype: str, :class:`~obspy.core.inventory.channel.Channel`
        :return: Network code, channel retrieved.
        """
        if self.inventory is None:
            return None

        sub_inventory = self.inventory.select(station=station, time=time)

        if len(sub_inventory) == 0:
            self._warn("Can't find station %s in inventory" % station)
            return None

        if len(sub_inventory) > 1:
            self._warn('Several stations %s have been found with different '
                       'network code in inventory' %
                       station)
            return None

        network = sub_inventory[0]

        sta = [cha for sta in network for cha in sta]

        # Select only vertical channels
        channels = [cha for cha in sta if cha.code[-1] == 'Z']

        if len(channels) == 0:
            self._warn("Can't find channel for station %s" % station)
            return network.code, None

        if len(channels) > 1:
            # Choose first location code
            location_codes = [channel.location_code for channel in channels]
            location_code = sorted(location_codes)[0]
            channels = [channel for channel in channels
                        if channel.location_code == location_code]

            codes = [channel.code for channel in channels]
            code = None
            if len(codes) == 1:
                code = codes[0]
            else:
                # Choose channel code by priority
                # HHZ > EHZ > ELZ > BHZ > LHZ > SHZ
                priority = ['HHZ', 'EHZ', 'ELZ', 'BHZ', 'LHZ', 'SHZ']
                code = next((code for code in priority if code in codes), None)

            if code is None:
                if len(network) > 1:
                    self._warn('Several stations %s or location code have '
                               'been found in inventory' % station)
                else:
                    self._warn('Several channels have been found for '
                               'station %s' % station)

                return network.code, None

            channel = next((channel for channel in channels
                           if channel.code == code))

            self._warn('Several stations, location codes or channels have '
                       'been found, choose %s.%s.%s.%s' %
                       (network.code, station, channel.location_code,
                        channel.code))
            return network.code, channel

        return network.code, channels[0]

    def _parse_event(self, first_line):
        """
        Parse an event.

        :type first_line: str
        :param first_line: First line of an event block, which contains
            the event id.
        :rtype: :class:`~obspy.core.event.event.Event`
        :return: The parsed event or None.
        """
        event_id = first_line[5:].strip()
        # Skip event without id
        if not event_id:
            self._warn('Missing event id')
            return None

        event = Event()

        origin, origin_res_id = self._parse_origin(event)
        # Skip event without origin
        if not origin:
            return None

        line = self._skip_empty_lines()

        self._parse_region_name(line, event)
        self._parse_arrivals(event, origin, origin_res_id)

        # Origin ResourceIdentifier should be set at the end, when
        # Arrivals are already set.
        origin.resource_id = origin_res_id
        event.origins.append(origin)

        event.preferred_origin_id = origin.resource_id.id

        # Must be done after the origin parsing
        event.creation_info = self._get_creation_info()

        public_id = "event/%s" % event_id
        event.resource_id = self._get_res_id(public_id)

        event.scope_resource_ids()

        return event

    def _parse_origin(self, event):
        """
        Parse an origin.

        :type event: :class:`~obspy.core.event.event.Event`
        :param event: Event of the origin.
        :rtype: :class:`~obspy.core.event.origin.Origin`,
            :class:`~obspy.core.event.resourceid.ResourceIdentifier`
        :returns: Parsed origin or None, resource identifier of the
            origin.
        """
        # Store magnitudes in a list to keep their positions.
        # Needed for the magnitude errors.
        magnitudes = []

        # Skip 2 lines of header
        next(self.lines)
        next(self.lines)

        line = self._skip_empty_lines()

        origin, origin_res_id = \
            self._parse_first_line_origin(line, event, magnitudes)

        # If some crucial information can't be parsed, return None
        if not origin:
            return None, None

        line = self._skip_empty_lines()

        # File from ldg can have author on several lines
        while re.match(r'\s{105,}\w+\s*', line):
            fields = self.fields['line_1']
            origin.creation_info.author += line[fields['author']].strip()
            line = self._skip_empty_lines()

        self._parse_second_line_origin(line, event, origin, magnitudes)

        # Magnitude resource_id can only be set after that magnitude
        # errors are set.
        for i in range(3):
            magnitude = magnitudes[i]
            if magnitude is not None:
                public_id = "magnitude/%s" % i
                mag_res_id = \
                    self._get_res_id(public_id, parent_res_id=origin_res_id)
                magnitude.resource_id = mag_res_id

                if event.preferred_magnitude_id is None:
                    event.preferred_magnitude_id = magnitude.resource_id.id

        return origin, origin_res_id

    def _parse_first_line_origin(self, line, event, magnitudes):
        """
        Parse the first line of origin data.

        :type line: str
        :param line: Line to parse.
        :type event: :class:`~obspy.core.event.event.Event`
        :param event: Event of the origin.
        :type magnitudes: list of
            :class:`~obspy.core.event.magnitude.Magnitude`
        :param magnitudes: Store magnitudes in a list to keep
            their positions.
        :rtype: :class:`~obspy.core.event.origin.Origin`,
            :class:`~obspy.core.event.resourceid.ResourceIdentifier`
        :returns: Parsed origin or None, resource identifier of the
            origin.
        """
        magnitude_types = []
        magnitude_values = []
        magnitude_station_counts = []

        fields = self.fields['line_1']

        time_origin = line[fields['time']].strip()
        time_fixed_flag = line[fields['time_fixf']].strip()
        latitude = line[fields['lat']].strip()
        longitude = line[fields['lon']].strip()
        epicenter_fixed_flag = line[fields['epicenter_fixf']].strip()
        depth = line[fields['depth']].strip()
        depth_fixed_flag = line[fields['depth_fixf']].strip()
        phase_count = line[fields['n_def']].strip()
        station_count = line[fields['n_sta']].strip()
        azimuthal_gap = line[fields['gap']].strip()
        magnitude_types.append(line[fields['mag_type_1']].strip())
        magnitude_values.append(line[fields['mag_1']].strip())
        magnitude_station_counts.append(line[fields['mag_n_sta_1']].strip())
        magnitude_types.append(line[fields['mag_type_2']].strip())
        magnitude_values.append(line[fields['mag_2']].strip())
        magnitude_station_counts.append(line[fields['mag_n_sta_2']].strip())
        magnitude_types.append(line[fields['mag_type_3']].strip())
        magnitude_values.append(line[fields['mag_3']].strip())
        magnitude_station_counts.append(line[fields['mag_n_sta_3']].strip())
        author = line[fields['author']].strip()
        origin_id = line[fields['id']].strip()

        origin = Origin()
        origin.quality = OriginQuality()

        try:
            origin.time = UTCDateTime(time_origin.replace('/', '-'))
            origin.latitude = float(latitude)
            origin.longitude = float(longitude)
        except (TypeError, ValueError):
            self._warn('Missing origin data, skipping event')
            return None, None

        origin.time_fixed = time_fixed_flag.lower() == 'f'
        origin.epicenter_fixed = epicenter_fixed_flag.lower() == 'f'

        try:
            # Convert value from km to m
            origin.depth = float(depth) * 1000
        except ValueError:
            pass
        try:
            origin.depth_type = DEPTH_TYPES[depth_fixed_flag]
        except KeyError:
            origin.depth_type = OriginDepthType('from location')
        try:
            origin.quality.used_phase_count = int(phase_count)
            origin.quality.associated_phase_count = int(phase_count)
        except ValueError:
            pass
        try:
            origin.quality.used_station_count = int(station_count)
            origin.quality.associated_station_count = int(station_count)
        except ValueError:
            pass
        try:
            origin.quality.azimuthal_gap = float(azimuthal_gap)
        except ValueError:
            pass

        self.author = author
        origin.creation_info = self._get_creation_info()

        public_id = "origin/%s" % origin_id
        origin_res_id = self._get_res_id(public_id)

        for i in range(3):
            try:
                magnitude = Magnitude()
                magnitude.creation_info = self._get_creation_info()
                magnitude.magnitude_type = magnitude_types[i]
                magnitude.mag = float(magnitude_values[i])
                magnitude.station_count = int(magnitude_station_counts[i])
                magnitude.origin_id = origin_res_id
                magnitudes.append(magnitude)
                event.magnitudes.append(magnitude)
            except ValueError:
                # Magnitude can be empty but we need to keep the
                # position between mag1, mag2 or mag3.
                magnitudes.append(None)

        return origin, origin_res_id

    def _find_magnitude_by_type(self, event, origin_res_id, magnitude_type):
        for mag in event.magnitudes:
            if mag.origin_id == origin_res_id \
                    and mag.magnitude_type == magnitude_type:
                return mag

    def _parse_second_line_origin(self, line, event, origin, magnitudes):
        magnitude_errors = []

        fields = self.fields['line_2']

        standard_error = line[fields['rms']].strip()
        time_uncertainty = line[fields['ot_error']].strip()
        max_horizontal_uncertainty = line[fields['s_major']].strip()
        min_horizontal_uncertainty = line[fields['s_minor']].strip()
        azimuth_max_horizontal_uncertainty = line[fields['az']].strip()
        depth_uncertainty = line[fields['depth_err']].strip()
        min_distance = line[fields['min_dist']].strip()
        max_distance = line[fields['max_dist']].strip()
        magnitude_errors.append(line[fields['mag_err_1']].strip())
        magnitude_errors.append(line[fields['mag_err_2']].strip())
        magnitude_errors.append(line[fields['mag_err_3']].strip())
        analysis_type = line[fields['antype']].strip().lower()
        location_method = line[fields['loctype']].strip().lower()
        event_type = line[fields['evtype']].strip().lower()

        try:
            origin.quality.standard_error = float(standard_error)
        except ValueError:
            pass

        try:
            origin.time_errors.uncertainty = float(time_uncertainty)
        except ValueError:
            pass

        try:
            uncertainty = OriginUncertainty()
            # Convert values from km to m
            min_value = float(min_horizontal_uncertainty) * 1000
            max_value = float(max_horizontal_uncertainty) * 1000
            azimuth_value = float(azimuth_max_horizontal_uncertainty)
            description = OriginUncertaintyDescription('uncertainty ellipse')

            uncertainty.min_horizontal_uncertainty = min_value
            uncertainty.max_horizontal_uncertainty = max_value
            uncertainty.azimuth_max_horizontal_uncertainty = azimuth_value
            uncertainty.preferred_description = description
            origin.origin_uncertainty = uncertainty
        except ValueError:
            pass

        try:
            # Convert value from km to m
            origin.depth_errors.uncertainty = float(depth_uncertainty) * 1000
        except ValueError:
            pass

        try:
            origin.quality.minimum_distance = float(min_distance)
            origin.quality.maximum_distance = float(max_distance)
        except ValueError:
            self._warn('Missing minimum/maximum distance')

        for i in range(2):
            try:
                mag_errors = magnitudes[i].mag_errors
                mag_errors.uncertainty = float(magnitude_errors[i])
            except (AttributeError, ValueError):
                pass

        # No match for 'g' (guess)
        # We map 'g' to 'manual' and create a comment for origin
        try:
            origin.evaluation_mode = EVALUATION_MODES[analysis_type]

            if analysis_type == 'g':
                # comment: 'GSE2.0:antype=g'
                text = 'GSE2.0:antype=g'
                comment = self._comment(text)
                origin.comments.append(comment)
        except KeyError:
            self._warn('Wrong analysis type')

        if location_method not in LOCATION_METHODS.keys():
            location_method = 'o'
        method = LOCATION_METHODS[location_method]
        method_id = "method/%s" % method
        origin.method_id = self._get_res_id(method_id)

        if event_type not in EVENT_TYPES.keys():
            event_type = 'uk'
            self._warn('Wrong or unknown event type')
        event_data = EVENT_TYPES[event_type]
        event.event_type_certainty, event.event_type = event_data

        # comment: 'GSE2.0:evtype=<evtype>'
        if event_type:
            text = 'GSE2.0:evtype=%s' % event_type
            comment = self._comment(text)
            event.comments.append(comment)

    def _parse_region_name(self, line, event):
        event_description = EventDescription()
        event_description.text = line.strip()
        event_description.type = EventDescriptionType('region name')
        event.event_descriptions.append(event_description)

    def _parse_arrivals(self, event, origin, origin_res_id):
        # Skip header of arrivals
        next(self.lines)

        # Stop the loop after 2 empty lines (according to the standard).
        previous_line_empty = False

        for line in self.lines:
            line_empty = not line or line.isspace()

            if not self.event_point_separator:
                # Event are separated by two empty lines
                if line_empty and previous_line_empty:
                    break
            else:
                # Event are separated by '.'
                if line.startswith('.'):
                    break

            previous_line_empty = line_empty

            if line_empty:
                # Skip empty lines when the loop should be stopped by
                # point
                continue

            magnitude_types = []
            magnitude_values = []

            fields = self.fields['arrival']

            station = line[fields['sta']].strip()
            distance = line[fields['dist']].strip()
            event_azimuth = line[fields['ev_az']].strip()
            evaluation_mode = line[fields['picktype']].strip()
            direction = line[fields['direction']].strip()
            onset = line[fields['detchar']].strip()
            phase = line[fields['phase']].strip()
            time = line[fields['time']].strip().replace('/', '-')
            time_residual = line[fields['t_res']].strip()
            arrival_azimuth = line[fields['azim']].strip()
            azimuth_residual = line[fields['az_res']].strip()
            slowness = line[fields['slow']].strip()
            slowness_residual = line[fields['s_res']].strip()
            time_defining_flag = line[fields['t_def']].strip()
            azimuth_defining_flag = line[fields['a_def']].strip()
            slowness_defining_flag = line[fields['s_def']].strip()
            snr = line[fields['snr']].strip()
            amplitude_value = line[fields['amp']].strip()
            period = line[fields['per']].strip()
            magnitude_types.append(line[fields['mag_type_1']].strip())
            magnitude_values.append(line[fields['mag_1']].strip())
            magnitude_types.append(line[fields['mag_type_2']].strip())
            magnitude_values.append(line[fields['mag_2']].strip())
            line_id = line[fields['id']].strip()

            # Don't take pick and arrival with wrong time residual
            if '*' in time_residual:
                continue

            try:
                pick = Pick()
                pick.creation_info = self._get_creation_info()
                pick.waveform_id = WaveformStreamID()
                pick.waveform_id.station_code = station
                pick.time = UTCDateTime(time)

                network_code = self.default_network_code
                location_code = self.default_location_code
                channel_code = self.default_channel_code

                try:
                    network_code, channel = self._get_channel(station,
                                                              pick.time)
                    if channel:
                        channel_code = channel.code
                        location_code = channel.location_code
                except TypeError:
                    pass

                pick.waveform_id.network_code = network_code
                pick.waveform_id.channel_code = channel_code
                if location_code:
                    pick.waveform_id.location_code = location_code

                try:
                    ev_mode = EVALUATION_MODES[evaluation_mode]
                    pick.evaluation_mode = ev_mode
                except KeyError:
                    pass
                try:
                    pick.polarity = PICK_POLARITIES[direction]
                except KeyError:
                    pass
                try:
                    pick.onset = PICK_ONSETS[onset]
                except KeyError:
                    pass
                pick.phase_hint = phase
                try:
                    pick.backazimuth = float(arrival_azimuth)
                except ValueError:
                    pass
                try:
                    pick.horizontal_slowness = float(slowness)
                except ValueError:
                    pass

                public_id = "pick/%s" % line_id
                pick.resource_id = self._get_res_id(public_id)
                event.picks.append(pick)
            except (TypeError, ValueError, AttributeError):
                # Can't parse pick, skip arrival and amplitude parsing
                continue

            arrival = Arrival()
            arrival.creation_info = self._get_creation_info()

            try:
                arrival.pick_id = pick.resource_id.id
            except AttributeError:
                pass
            arrival.phase = phase
            try:
                arrival.azimuth = float(event_azimuth)
            except ValueError:
                pass
            try:
                arrival.distance = float(distance)
            except ValueError:
                pass
            try:
                arrival.time_residual = float(time_residual)
            except ValueError:
                pass
            try:
                arrival.backazimuth_residual = float(azimuth_residual)
            except ValueError:
                pass
            try:
                arrival.horizontal_slowness_residual = float(slowness_residual)
            except ValueError:
                pass

            if time_defining_flag == 'T':
                arrival.time_weight = 1

            if azimuth_defining_flag == 'A':
                arrival.backazimuth_weight = 1

            if slowness_defining_flag == 'S':
                arrival.horizontal_slowness_weight = 1

            public_id = "arrival/%s" % line_id
            arrival.resource_id = self._get_res_id(public_id,
                                                   parent_res_id=origin_res_id)
            origin.arrivals.append(arrival)

            try:
                amplitude = Amplitude()
                amplitude.creation_info = self._get_creation_info()
                amplitude.generic_amplitude = float(amplitude_value)
                try:
                    amplitude.pick_id = pick.resource_id
                    amplitude.waveform_id = pick.waveform_id
                except AttributeError:
                    pass
                try:
                    amplitude.period = float(period)
                except ValueError:
                    pass
                try:
                    amplitude.snr = float(snr)
                except ValueError:
                    pass

                for i in [0, 1]:
                    if magnitude_types[i] and not magnitude_types[i].isspace():
                        amplitude.magnitude_hint = magnitude_types[i]

                public_id = "amplitude/%s" % line_id
                amplitude.resource_id = self._get_res_id(public_id)
                event.amplitudes.append(amplitude)
                for i in [0, 1]:
                    sta_mag = StationMagnitude()
                    sta_mag.creation_info = self._get_creation_info()
                    sta_mag.origin_id = origin_res_id
                    sta_mag.amplitude_id = amplitude.resource_id
                    sta_mag.station_magnitude_type = magnitude_types[i]
                    try:
                        sta_mag.mag = magnitude_values[i]
                    except ValueError:
                        continue
                    sta_mag.waveform_id = pick.waveform_id
                    public_id = "magnitude/station/%s/%s" % (line_id, i)
                    sta_mag.resource_id = self._get_res_id(public_id)
                    event.station_magnitudes.append(sta_mag)

                    # Associate station mag with network mag of same type
                    mag = self._find_magnitude_by_type(event, origin_res_id,
                                                       magnitude_types[i])
                    if mag:
                        contrib = StationMagnitudeContribution()
                        contrib.station_magnitude_id = sta_mag.resource_id
                        contrib.weight = 1.0
                        mag.station_magnitude_contributions.append(contrib)
            except ValueError:
                pass

    def _deserialize(self):
        catalog = Catalog()
        catalog.description = 'Created from GSE2 format'
        catalog.creation_info = self._get_creation_info()

        # Flag used to ignore line which aren't in a BEGIN-STOP block
        begin_block = False
        # Flag used to ignore line which aren't in a BULLETIN block
        bulletin_block = False

        try:
            for line in self.lines:
                if line.startswith('BEGIN'):
                    if begin_block:
                        # 2 BEGIN without STOP
                        message = self._add_line_nb('Missing STOP tag')
                        raise GSE2BulletinSyntaxError(message)
                    else:
                        # Enter a BEGIN block
                        begin_block = True

                    self._check_header(line)
                elif line.startswith('STOP'):
                    if begin_block:
                        # Exit a BEGIN-STOP block
                        begin_block = False
                    else:
                        # STOP without BEGIN
                        message = self._add_line_nb('Missing BEGIN tag')
                        raise GSE2BulletinSyntaxError(message)
                elif line.startswith('DATA_TYPE'):
                    bulletin_block = line[10:18] == 'BULLETIN'

                if not begin_block or not bulletin_block:
                    # Not in a BEGIN-STOP block, nor a DATA_TYPE BULLETIN
                    # block.
                    continue

                # If a "Reviewed Event Bulletin" or "Reviewed Bulletin"
                # line exists, put it in comment
                if 'Reviewed Event Bulletin' in line \
                        or 'Reviewed Bulletin' in line:
                    comment = self._comment(line.strip())
                    if comment.text:
                        catalog.comments.append(comment)
                # Detect start of an event
                elif line.startswith('EVENT'):
                    event = self._parse_event(line)
                    if event:
                        catalog.append(event)

        except StopIteration:
            message = self._add_line_nb('Unexpected EOF while parsing')
            raise GSE2BulletinSyntaxError(message)
        except Exception:
            self._warn('Unexpected error')
            raise

        if begin_block:
            # BEGIN-STOP block not closed
            text = 'Unexpected EOF while parsing, BEGIN-STOP block not closed'
            message = self._add_line_nb(text)
            raise GSE2BulletinSyntaxError(message)

        catalog.resource_id = self._get_res_id('event/evid')

        return catalog


def _read_gse2(filename, inventory=None, default_network_code='XX',
               default_location_code=None, default_channel_code=None,
               res_id_prefix='smi:local', fields=None,
               event_point_separator=False, agency=None):
    """
    Read a GSE2.0 bulletin file and returns an ObsPy Catalog object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.event.read_events` function, call this
        instead.

    :type filename: str
    :param filename: File or file-like object in text mode.
    :type inventory: :class:`~obspy.core.inventory.inventory.Inventory`
    :param inventory: Inventory used to retrieve network code, location code
        and channel code of stations.
    :type default_network_code: str
    :param default_network_code: Default network code used if stations
        are not found in the inventory.
    :type default_location_code: str
    :param default_location_code: Location code used if stations are
        not found in the inventory.
    :type default_channel_code: str
    :param default_channel_code: Default channel code used if stations
        are not found in the inventory.
    :type res_id_prefix: str
    :param res_id_prefix: Prefix used
        in :class:`~obspy.core.event.resourceid.ResourceIdentifier` attributes.
    :type fields: dict
    :param fields: dictionary of positions of input fields, used if input file
        is non-standard
    :type event_point_separator: bool
    :param event_point_separator: ``True`` if events are separated by
        point rather than 2 empty lines.
    :type agency: str
    :param agency: Agency that generated the file.
    :rtype: :class:`~obspy.core.event.Catalog`
    :return: An ObsPy Catalog object.

    .. rubric:: Example

    >>> from obspy import read_events
    >>> default_fields = {
    ...     'line_1': {
    ...         'time': slice(0, 21),
    ...         'time_fixf': slice(22, 23),
    ...         'lat': slice(25, 33),
    ...         'lon': slice(34, 43),
    ...         'epicenter_fixf': slice(44, 45),
    ...         'depth': slice(47, 52),
    ...         'depth_fixf': slice(53, 54),
    ...         'n_def': slice(56, 60),
    ...         'n_sta': slice(61, 65),
    ...         'gap': slice(66, 69),
    ...         'mag_type_1': slice(71, 73),
    ...         'mag_1': slice(73, 77),
    ...         'mag_n_sta_1': slice(78, 80),
    ...         'mag_type_2': slice(82, 84),
    ...         'mag_2': slice(84, 88),
    ...         'mag_n_sta_2': slice(89, 91),
    ...         'mag_type_3': slice(93, 95),
    ...         'mag_3': slice(95, 99),
    ...         'mag_n_sta_3': slice(100, 102),
    ...         'author': slice(104, 112),
    ...         'id': slice(114, 122),
    ...     },
    ...     'line_2': {
    ...         'rms': slice(5, 10),
    ...         'ot_error': slice(15, 21),
    ...         's_major': slice(25, 31),
    ...         's_minor': slice(32, 38),
    ...         'az': slice(40, 43),
    ...         'depth_err': slice(49, 54),
    ...         'min_dist': slice(56, 62),
    ...         'max_dist': slice(63, 69),
    ...         'mag_err_1': slice(74, 77),
    ...         'mag_err_2': slice(85, 88),
    ...         'mag_err_3': slice(96, 99),
    ...         'antype': slice(104, 105),
    ...         'loctype': slice(106, 107),
    ...         'evtype': slice(108, 110),
    ...     },
    ...     'arrival': {
    ...         'sta': slice(0, 5),
    ...         'dist': slice(6, 12),
    ...         'ev_az': slice(13, 18),
    ...         'picktype': slice(19, 20),
    ...         'direction': slice(20, 21),
    ...         'detchar': slice(21, 22),
    ...         'phase': slice(23, 30),
    ...         'time': slice(31, 52),
    ...         't_res': slice(53, 58),
    ...         'azim': slice(59, 64),
    ...         'az_res': slice(65, 71),
    ...         'slow': slice(72, 77),
    ...         's_res': slice(78, 83),
    ...         't_def': slice(84, 85),
    ...         'a_def': slice(85, 86),
    ...         's_def': slice(86, 87),
    ...         'snr': slice(88, 93),
    ...         'amp': slice(94, 103),
    ...         'per': slice(104, 109),
    ...         'mag_type_1': slice(110, 112),
    ...         'mag_1': slice(112, 116),
    ...         'mag_type_2': slice(117, 119),
    ...         'mag_2': slice(119, 123),
    ...         'id': slice(124, 132),
    ...     },
    ... }
    >>> # Only non-standard field indexes are required
    >>> fields = {
    ...     'line_1': {
    ...         'author': slice(105, 113),
    ...         'id': slice(114, 123),
    ...     },
    ...     'line_2': {
    ...         'az': slice(40, 46),
    ...         'antype': slice(105, 106),
    ...         'loctype': slice(107, 108),
    ...         'evtype': slice(109, 111),
    ...     },
    ...     'arrival': {
    ...         'amp': slice(94, 104),
    ...     },
    ... }
    >>> catalog = read_events('/path/to/bulletin/gse_2.0_non_standard.txt',
    ... default_network_code='FR', res_id_prefix='quakeml:abc',
    ... fields=fields, event_point_separator=True)
    >>> print(catalog)
    2 Event(s) in Catalog:
    1995-01-16T07:26:52.400000Z | +39.450,  +20.440 | 3.6  mb | manual
    1995-01-16T07:27:07.300000Z | +50.772, -129.760 | 1.2  Ml | manual
    """
    return Unpickler(inventory, default_network_code, default_location_code,
                     default_channel_code, res_id_prefix, fields,
                     event_point_separator, agency).load(filename)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np

from obspy.taup.velocity_model import VelocityModel


class TestTauPyVelocityModel:
    def test_read_velocity_model(self, testdata):
        for filename in ['iasp91.tvel', 'iasp91_w_comment.tvel', 'iasp91.nd',
                         'iasp91_w_comment.nd']:
            velocity_model = testdata[filename]
            test2 = VelocityModel.read_velocity_file(velocity_model)

            assert len(test2.layers) == 129
            assert len(test2) == 129

            assert test2.radius_of_planet == 6371.0
            assert test2.moho_depth == 35
            assert test2.cmb_depth == 2889.0
            assert test2.iocb_depth == 5153.9
            assert test2.min_radius == 0.0
            assert test2.max_radius == 6371.0

            assert test2.validate()

            np.testing.assert_equal(
                test2.get_discontinuity_depths(),
                [0.0, 20.0, 35.0, 210.0, 410.0, 660.0, 2889.0, 5153.9, 6371.0])

            # check boundary cases
            assert test2.layer_number_above(6371) == 128
            assert test2.layer_number_below(0) == 0

            # evaluate at cmb
            assert test2.evaluate_above(2889.0, 'p') == 13.6908
            assert test2.evaluate_below(2889.0, 'D') == 9.9145
            assert test2.depth_at_top(50) == 2393.5
            assert test2.depth_at_bottom(50) == 2443.0
            assert not test2.fix_discontinuity_depths()



# -*- coding: utf-8 -*-
"""
The obspy.clients.seedlink.slclient test suite.
"""

import pytest

from obspy import UTCDateTime
from obspy.clients.seedlink.slclient import SLClient


pytestmark = pytest.mark.network


class TestSLClient():
    """
    These test routines use SLClient, which is actually not expected to be
    used, but expected to be subclassed since at least the packet_handler
    method needs to be reimplemented.
    The original packet_handler method does not do anything with the received
    metadata or waveforms but only prints some information and not even checks
    when the requested waveform data (if any) is fully received to break out of
    the loop that is waiting for more packets. That is why these data requests
    below get stuck and not terminate.
    """

    @pytest.mark.skipif(
        __name__ != '__main__', reason='test must be started manually')
    def test_info(self):
        sl_client = SLClient()
        sl_client.slconn.set_sl_address("geofon.gfz.de:18000")
        sl_client.infolevel = "ID"
        sl_client.verbose = 2
        sl_client.initialize()
        sl_client.run()

    @pytest.mark.skipif(
        __name__ != '__main__', reason='test must be started manually')
    def test_time_window(self):
        sl_client = SLClient()
        sl_client.slconn.set_sl_address("geofon.gfz.de:18000")
        sl_client.multiselect = ("GE_STU:BHZ")
        # set a time window from 2 min - 1 min in the past
        dt = UTCDateTime()
        sl_client.begin_time = (dt - 120.0).format_seedlink()
        sl_client.end_time = (dt - 60.0).format_seedlink()
        sl_client.verbose = 2
        sl_client.initialize()
        sl_client.run()

    @pytest.mark.skipif(
        __name__ != '__main__', reason='test must be started manually')
    def test_issue708(self):
        sl_client = SLClient()
        sl_client.slconn.set_sl_address("rtserve.iris.washington.edu:18000")
        sl_client.multiselect = ("G_FDFM:00BHZ, G_SSB:00BHZ")
        # set a time window from 2 min - 1 min in the past
        dt = UTCDateTime()
        sl_client.begin_time = (dt - 120.0).format_seedlink()
        sl_client.end_time = (dt - 60.0).format_seedlink()
        sl_client.verbose = 2
        sl_client.initialize()
        sl_client.run()
